#!/usr/bin/python
# =============================================================================
#
#  ######     ###    ########  ##    ##  ######   #######  ##       
# ##    ##   ## ##   ##     ##  ##  ##  ##    ## ##     ## ##       
# ##        ##   ##  ##     ##   ####   ##       ##     ## ##       
#  ######  ##     ## ########     ##     ######  ##     ## ##       
#       ## ######### ##           ##          ## ##     ## ##       
# ##    ## ##     ## ##           ##    ##    ## ##     ## ##       
#  ######  ##     ## ##           ##     ######   #######  ########
#
# =============================================================================
#
# SuperArmor's Python Solana library.
# (c) SuperArmor
#
# module: mass token sell (dump) via Jupiter
#
# =============================================================================
# 
from   solana.rpc.api    import Client, Pubkey, Keypair
from   solana.exceptions import SolanaRpcException
from   typing            import List, Union
from   queue             import Queue, Empty
from ..helpers           import MakePubkey, SapysolPubkey
from ..token             import SapysolToken
from ..jupag             import SapysolJupagParams, SapysolJupag
from ..tx                import SapysolTxParams, SapysolTxStatus, SapysolTx, SendAndWaitBatchTx
from  .batcher           import SapysolBatcher
import logging

# =============================================================================
# 
class SapysolTokenSelloff:
    def __init__(self,
                 connection:         Client,
                 walletsList:        List[Keypair],
                 tokenToSell:        SapysolPubkey,
                 tokenToBuy:         SapysolPubkey,
                 txParams:           SapysolTxParams          = SapysolTxParams(),
                 swapParams:         SapysolJupagParams       = SapysolJupagParams(),
                 connectionOverride: List[Union[str, Client]] = None,
                 numThreads:         int = 10):

        assert(all(isinstance(n, Keypair) for n in walletsList))
        self.CONNECTION:          Client                   = connection
        self.TOKEN_TO_SELL:       SapysolToken             = SapysolToken(connection=connection, tokenMint=MakePubkey(tokenToSell))
        self.TOKEN_TO_BUY:        SapysolToken             = SapysolToken(connection=connection, tokenMint=MakePubkey(tokenToBuy ))
        self.TX_PARAMS:           SapysolTxParams          = txParams
        self.SWAP_PARAMS:         SapysolJupagParams       = swapParams
        self.CONNECTION_OVERRIDE: List[Union[str, Client]] = connectionOverride
        self.BATCHER:             SapysolBatcher           = SapysolBatcher(callback    = self.SellSingle,
                                                                            entityList  = walletsList,
                                                                            entityKwarg = "wallet",
                                                                            numThreads  = numThreads)

    # ========================================
    #
    def SellSingle(self, wallet: Keypair):
        while True:
            balance:   int = self.TOKEN_TO_SELL.GetWalletBalanceLamports(walletAddress=wallet.pubkey())
            delimiter: int = 10**self.TOKEN_TO_SELL.TOKEN_INFO.decimals
            if balance <= 0:
                logging.info(f"Wallet: {str(wallet.pubkey()):>44}; balance: 0, skipping...")
                break
            else:
                logging.info(f"Wallet: {str(wallet.pubkey()):>44}; balance: {balance / delimiter}, trying to sell all...")

            quote = SapysolJupag.GetSwapQuote(connection = self.CONNECTION,
                                              tokenFrom  = self.TOKEN_TO_SELL.TOKEN_MINT,
                                              tokenTo    = self.TOKEN_TO_BUY.TOKEN_MINT,
                                              inAmount   = balance,
                                              swapParams = self.SWAP_PARAMS)
            txb64         = SapysolJupag.GetSwapTxBase64(walletAddress=wallet.pubkey(), coinQuote=quote, swapParams=self.SWAP_PARAMS)
            tx: SapysolTx = SapysolTx(connection=self.CONNECTION, payer=wallet, txParams=self.TX_PARAMS)
            tx.FromBase64(b64=txb64)
            result: SapysolTxStatus = tx.Sign([wallet]).SendAndWait(self.CONNECTION_OVERRIDE)
            if result == SapysolTxStatus.SUCCESS:
                break

    # ========================================
    #
    def Start(self, **kwargs) -> None:
        self.BATCHER.Start(**kwargs)

# =============================================================================
# 



#!/usr/bin/python
# =============================================================================
#
#  ######     ###    ########  ##    ##  ######   #######  ##       
# ##    ##   ## ##   ##     ##  ##  ##  ##    ## ##     ## ##       
# ##        ##   ##  ##     ##   ####   ##       ##     ## ##       
#  ######  ##     ## ########     ##     ######  ##     ## ##       
#       ## ######### ##           ##          ## ##     ## ##       
# ##    ## ##     ## ##           ##    ##    ## ##     ## ##       
#  ######  ##     ## ##           ##     ######   #######  ########
#
# =============================================================================
#
# SuperArmor's Python Solana library.
# (c) SuperArmor
#
# module: helpers
#
# =============================================================================
# 
from   solana.rpc.api        import Client, Pubkey, Keypair, Commitment
from   solana.rpc.commitment import Commitment
from   solders.account       import Account, AccountJSON
from   typing                import List, Any, Union
from   pybip39               import Mnemonic, Seed
import logging
import json
import os

# ================================================================================
#
LAMPORTS_PER_SOL:          int    = 1_000_000_000
METADATA_PROGRAM_ID:       Pubkey = Pubkey.from_string("metaqbxxUerdq28cj1RbAWkYQm3ybzjb6a8bt518x1s" )
SYSTEM_PROGRAM_ID:         Pubkey = Pubkey.from_string("11111111111111111111111111111111"            )
SYSVAR_RENT_PUBKEY:        Pubkey = Pubkey.from_string("SysvarRent111111111111111111111111111111111" )
SYSVAR_CLOCK_PUBKEY:       Pubkey = Pubkey.from_string("SysvarC1ock11111111111111111111111111111111" )
SYSVAR_SLOT_HASHES_PUBKEY: Pubkey = Pubkey.from_string("SysvarS1otHashes111111111111111111111111111" )
TOKEN_2022_PROGRAM_ID:     Pubkey = Pubkey.from_string("TokenzQdBNbLqP5VEhdkAS6EPFLC1PHnBqCXEpPxuEb" )
# Address of the special mint for wrapped native SOL in spl-token-2022 */
NATIVE_MINT_2022:          Pubkey = Pubkey.from_string("9pan9bMn5HatX4EJdBwg9VgCa7Uz5HL8N1m5D3NdXejP")

# ================================================================================
# Create path if it doesn't exist
#
def EnsurePathExists(path: str):
    if path == "":
        return
    if not os.path.exists(path):
        os.makedirs(path, exist_ok = True)

# ================================================================================
#
def SetupLogging(fileName: str = "log.log",
                 format:   str = "%(name)s: %(asctime)s | %(levelname)s | %(filename)s:%(lineno)s | %(process)d >>> %(message)s",
                 dateFmt:  str = "%Y-%m-%dT%H:%M:%SZ",
                 logLevel: int = logging.INFO):
    EnsurePathExists(os.path.dirname(fileName))
    logging.basicConfig(filename=fileName, filemode="a", format=format, datefmt=dateFmt, level=logLevel)
    logging.getLogger().addHandler(logging.StreamHandler())

# ================================================================================
#
def NestedAttributeExists(target: object, attributePath: str) -> bool:
    parts = attributePath.split(".")
    while parts:
       next, parts = parts[0], parts[1:]
       target = getattr(target, next)
       if target is None:
           return False
    return True

# ================================================================================
# 
def GetModulePath(fileName = __file__) -> str:
    return os.path.dirname(os.path.realpath(fileName))

# ================================================================================
# 
def ListToChunks(baseList: List[Any], chunkSize: int) -> List[List[Any]]:
    result = []
    chunks = baseList
    while chunks:
        chunk, chunks = chunks[:chunkSize], chunks[chunkSize:]
        result.append(chunk)
    return result

# ================================================================================
# Either Pubkey string or Pubkey or anything that can be a Pubkey
#
SapysolPubkey = Union[str, bytes, Keypair, Pubkey]
#
def MakePubkey(pubkey: SapysolPubkey) -> Pubkey:
    if pubkey is None:
        return None

    if isinstance(pubkey, Pubkey):
        return pubkey
    if isinstance(pubkey, Keypair):
        return pubkey.pubkey()
    elif isinstance(pubkey, bytes):
        return Pubkey.from_bytes(pubkey)
    elif isinstance(pubkey, str):
        try: 
            return Pubkey.from_string(pubkey)
        except:
            return Pubkey.from_json(pubkey)
    return None

# ================================================================================
# Either Keypair JSON file path or Keypair or anything that can be a Keypair
#
SapysolKeypair = Union[str, bytes, Keypair]
#
def MakeKeypair(keypair: SapysolKeypair) -> Keypair:
    if keypair is None:
        return None

    if isinstance(keypair, Keypair):
        return keypair
    elif isinstance(keypair, bytes):
        return Keypair.from_bytes(keypair)
    elif isinstance(keypair, str):
        # First try to load file
        try:
            with open(keypair) as f:
                result = json.loads(f.read())
                return Keypair.from_bytes(result)
        except:
            pass
        # Second - maybe it is a mnemonic?
        try:
            mnemonic = Mnemonic.from_phrase(keypair)
            seed     = Seed(mnemonic=mnemonic, password="")
            return Keypair.from_seed_and_derivation_path(seed=bytes(seed), dpath="m/44'/501'/0'/0'")
        except:
            pass
        # Third - maybe it is already a json?
        try: 
            return Keypair.from_json(keypair)
        except:
            pass
    return None

# ================================================================================
#
def GetFilesFromPath(path: str, endsWith: str=".json") -> List[str]:
    files = os.listdir(path)
    return [os.path.join(dir, f) for f in files if os.path.isfile(os.path.join(dir, f)) and f.lower().endswith(endsWith)]

def GetKeypairsFromPath(path: str, endsWith: str=".json") -> List[Keypair]:
    files    = GetFilesFromPath(path=path, endsWith=endsWith)
    keypairs = [MakeKeypair(f) for f in files]
    return keypairs

def GetPubkeysFromKeypairs(keypairList: List[SapysolKeypair]) -> List[Pubkey]:
    return [ MakeKeypair(x).pubkey() for x in keypairList ]

# ================================================================================
#
def FetchAccounts(connection:    Client, 
                  pubkeys:       List[SapysolPubkey],
                  chunkSize:     int           = 100,
                  requiredOwner: SapysolPubkey = None,
                  commitment:    Commitment    = None,
                  parseToJson:   bool          = False) -> Union[List[Account], List[AccountJSON]]:
    results = []
    _pubkeys: List[Pubkey]       = [MakePubkey(pk) for pk in pubkeys]
    chunks:   List[List[Pubkey]] = ListToChunks(baseList=_pubkeys, chunkSize=chunkSize)
    func = connection.get_multiple_accounts_json_parsed if parseToJson else connection.get_multiple_accounts
    for chunk in chunks:
        entries: List[Account] = func(pubkeys=chunk, commitment=commitment).value
        if requiredOwner is not None and not all(e is None or (e.owner==requiredOwner) for e in entries):
            raise ValueError("Account does not belong to this program!")
        results += entries
    return results

# ================================================================================
#
def FetchAccount(connection:    Client, 
                 pubkey:        SapysolPubkey,
                 requiredOwner: SapysolPubkey = None,
                 commitment:    Commitment    = None,
                 parseToJson:   bool          = False) -> Union[Account, AccountJSON]:
    return FetchAccounts(connection    = connection,
                         pubkeys       = [pubkey],
                         requiredOwner = requiredOwner,
                         commitment    = commitment,
                         parseToJson   = parseToJson)[0]


# ===============================================================================
# 
def DivmodJsBignumber(dividend, divisor) -> tuple[int, int]:
    """
    Performs division to replicate JavaScript's division behavior accurately,
    with quotient rounding towards zero and appropriate remainder calculation.
    
    :param dividend: The number to be divided.
    :param divisor: The number by which to divide.
    :return: A tuple (quotient, remainder) accurately mimicking JavaScript behavior.
    """
    if divisor == 0:
        raise ValueError("Divisor cannot be zero.")
    
    quotient  = dividend // divisor
    remainder = dividend %  divisor
    
    # Ensure remainder has the same sign as the divisor
    if remainder != 0 and (dividend < 0) != (divisor < 0):
        quotient += 1
        remainder = remainder - divisor

    return quotient, remainder

# ================================================================================
#
# TODO: arweave https://github.com/Irys-xyz/gasless-uploader/blob/master/app/api/lazyFundSOL/route.ts


#!/usr/bin/python
# =============================================================================
#
#  ######     ###    ########  ##    ##  ######   #######  ##       
# ##    ##   ## ##   ##     ##  ##  ##  ##    ## ##     ## ##       
# ##        ##   ##  ##     ##   ####   ##       ##     ## ##       
#  ######  ##     ## ########     ##     ######  ##     ## ##       
#       ## ######### ##           ##          ## ##     ## ##       
# ##    ## ##     ## ##           ##    ##    ## ##     ## ##       
#  ######  ##     ## ##           ##     ######   #######  ########
#
# =============================================================================
#
# SuperArmor's Python Solana library.
# (c) SuperArmor
#
# module: mass wallets token balance
#
# =============================================================================
# 
from   solana.rpc.api import Client, Pubkey, Keypair
from   typing         import List, Union
from   queue          import Queue, Empty
from ..helpers        import MakePubkey, SapysolPubkey
from ..token          import SapysolToken
from  .batcher        import SapysolBatcher
import time
import threading
import logging

# =============================================================================
# 
class SapysolWalletsBalance:
    def __init__(self,
                 connection:  Client,
                 pubkeysList: List[Pubkey],
                 tokenMint:   SapysolPubkey,
                 numThreads:  int  = 50):

        assert(all(isinstance(n, Pubkey) for n in pubkeysList))
        self.CONNECTION:   Client         = connection
        self.PUBKEYS_LIST: List[Pubkey]   = pubkeysList
        self.TOKEN:        SapysolToken   = SapysolToken(connection=connection, tokenMint=MakePubkey(tokenMint))
        self.SOL_MINT:     Pubkey         = MakePubkey("So11111111111111111111111111111111111111112")
        self.RESULTS:      dict           = {}
        self.MUTEX:        threading.Lock = threading.Lock()
        self.BATCHER:      SapysolBatcher = SapysolBatcher(callback    = self.CheckSingle,
                                                           entityList  = pubkeysList,
                                                           entityKwarg = "walletAddress",
                                                           numThreads  = numThreads)

    # ========================================
    #
    def CheckSingle(self, walletAddress: Pubkey):
        balance: int = self.TOKEN.GetWalletBalanceLamports(walletAddress=walletAddress)

        # Include SOL when we check WSOL
        if self.TOKEN.TOKEN_MINT == self.SOL_MINT:
            balance += self.CONNECTION.get_balance(pubkey=walletAddress).value

        with self.MUTEX:
            self.RESULTS[walletAddress] = balance

    # ========================================
    #
    def Start(self, **kwargs) -> dict:
        self.RESULTS = {}
        self.BATCHER.Start(**kwargs)
        return self.RESULTS

    # ========================================
    #
    def OutputPretty(self, 
                     ignoreEmpty:       bool = False,
                     balanceInLamports: bool = False) -> None:

        delimiter = 10**self.TOKEN.TOKEN_INFO.decimals
        # HEADER
        logging.info((2+44+3+32+2)*"-")
        logging.info(f"| {'WALLET':<44} | {'BALANCE':<32} |")
        logging.info((2+44+3+32+2)*"-")

        sum:          int = 0
        walletsFull:  int = 0
        walletsEmpty: int = 0
        for wallet, balance in self.RESULTS.items():
            sum += balance
            if balance > 0:
                walletsFull  += 1
            else:
                walletsEmpty += 1

            if balanceInLamports:
                if not ignoreEmpty or balance > 0:
                    logging.info(f"| {str(wallet):>44} | {balance:<32} |")
            else:
                if not ignoreEmpty or balance > 0:
                    logging.info(f"| {str(wallet):>44} | {balance/delimiter:<32} |")

        # FOOTER
        delimiter: int = 10**self.TOKEN.TOKEN_INFO.decimals
        logging.info((2+44+3+32+2)*"-")

        logging.info(f"Wallets with balance: {walletsFull}" )
        logging.info(f"Wallets empty:        {walletsEmpty}")
        if balanceInLamports:
            logging.info(f"TOKENS TOTAL (lamports): {sum}\n")
        else:
            logging.info(f"TOKENS TOTAL: {sum/delimiter}\n")

# =============================================================================
# 



#!/usr/bin/python
# =============================================================================
#
#  ######     ###    ########  ##    ##  ######   #######  ##       
# ##    ##   ## ##   ##     ##  ##  ##  ##    ## ##     ## ##       
# ##        ##   ##  ##     ##   ####   ##       ##     ## ##       
#  ######  ##     ## ########     ##     ######  ##     ## ##       
#       ## ######### ##           ##          ## ##     ## ##       
# ##    ## ##     ## ##           ##    ##    ## ##     ## ##       
#  ######  ##     ## ##           ##     ######   #######  ########
#
# =============================================================================
#
# SuperArmor's Python Solana library.
# (c) SuperArmor
#
# module: SysvarClock
# https://docs.solanalabs.com/runtime/sysvars
#
# =============================================================================
# 
import typing
from   dataclasses           import dataclass
from   solana.rpc.api        import Client
from   solana.rpc.commitment import Commitment
import borsh_construct       as borsh
from ..helpers               import SYSVAR_CLOCK_PUBKEY, FetchAccount
#from   solders.clock  import Clock as SysvarClock
#from   solders.sysvar import CLOCK as SYSVAR_CLOCK_PUBKEY

# =============================================================================
# 
class SysvarClockJSON(typing.TypedDict):
    slot:                  int
    epoch_start_timestamp: int
    epoch:                 int
    leader_schedule_epoch: int
    unix_timestamp:        int

# =============================================================================
# 
@dataclass
class SysvarClock:
    layout: typing.ClassVar = borsh.CStruct(
        "slot"                  / borsh.U64,
        "epoch_start_timestamp" / borsh.I64,
        "epoch"                 / borsh.U64,
        "leader_schedule_epoch" / borsh.U64,
        "unix_timestamp"        / borsh.I64,
    )
    slot:                  int
    epoch_start_timestamp: int
    epoch:                 int
    leader_schedule_epoch: int
    unix_timestamp:        int

    # ========================================
    #
    @classmethod
    def fetch(cls,
              conn:       Client,
              commitment: typing.Optional[Commitment] = None) -> typing.Optional["SysvarClock"]:

        resp = FetchAccount(connection    = conn, 
                            pubkey        = SYSVAR_CLOCK_PUBKEY,
                            commitment    = commitment)
        return None if resp is None else cls.decode(resp.data)

    # ========================================
    #
    @classmethod
    def decode(cls, data: bytes) -> "SysvarClock":
        dec = SysvarClock.layout.parse(data)
        return cls(slot                  = dec.slot,
                   epoch_start_timestamp = dec.epoch_start_timestamp,
                   epoch                 = dec.epoch,
                   leader_schedule_epoch = dec.leader_schedule_epoch,
                   unix_timestamp        = dec.unix_timestamp)

    # ========================================
    #
    def to_json(self) -> SysvarClockJSON:
        return {
            "slot":                  self.slot,
            "epoch_start_timestamp": self.epoch_start_timestamp,
            "epoch":                 self.epoch,
            "leader_schedule_epoch": self.leader_schedule_epoch,
            "unix_timestamp":        self.unix_timestamp,
        }

    # ========================================
    #
    @classmethod
    def from_json(cls, obj: SysvarClockJSON) -> "SysvarClock":
        return cls(
            slot                  = obj["slot"],
            epoch_start_timestamp = obj["epoch_start_timestamp"],
            epoch                 = obj["epoch"],
            leader_schedule_epoch = obj["leader_schedule_epoch"],
            unix_timestamp        = obj["unix_timestamp"],
        )

# =============================================================================
# 



#!/usr/bin/python
# =============================================================================
#
#  ######     ###    ########  ##    ##  ######   #######  ##       
# ##    ##   ## ##   ##     ##  ##  ##  ##    ## ##     ## ##       
# ##        ##   ##  ##     ##   ####   ##       ##     ## ##       
#  ######  ##     ## ########     ##     ######  ##     ## ##       
#       ## ######### ##           ##          ## ##     ## ##       
# ##    ## ##     ## ##           ##    ##    ## ##     ## ##       
#  ######  ##     ## ##           ##     ######   #######  ########
#
# =============================================================================
#
# SuperArmor's Python Solana library.
# (c) SuperArmor
#
# module: jupag
#
# =============================================================================
# 
from  .helpers        import *
from  .token_cache    import TokenCacheEntry, TokenCache
from   solana.rpc.api import Client, Pubkey, Keypair
from   typing         import List, Any, TypedDict, Union, Literal
from   dataclasses    import dataclass
import logging
import requests

# =============================================================================
# 
@dataclass
class SapysolJupagParams:
    # Quote
    swapMode:                Literal["ExactIn", "ExactOut"] = "ExactIn"
    slippageBps:             int  = 50 # 50 = 0.5%
    onlyDirectRoutes:        bool = False
    asLegacyTransaction:     bool = False
    # Tx
    quoteAutoMultiplier:     int  = 1
    quotePrioFeeLamports:    str  = "auto"
    wrapAndUnwrapSol:        bool = True
    dynamicComputeUnitLimit: bool = True

# =============================================================================
# 
class SapysolJupag:
    # ========================================
    # 
    @staticmethod
    def GetSwapQuote(connection:          Client,
                     tokenFrom:           str, 
                     tokenTo:             str, 
                     inAmount:            float, 
                     desiredOutAmount:    float = None,
                     inAmountInLamports:  bool  = True, 
                     outAmountInLamports: bool  = True, 
                     swapParams:          SapysolJupagParams = SapysolJupagParams()
                    ) -> dict:
        FROM: TokenCacheEntry = TokenCache.GetToken(connection=connection, tokenMint=tokenFrom)
        TO:   TokenCacheEntry = TokenCache.GetToken(connection=connection, tokenMint=tokenTo  )

        finalInAmount = int(inAmount) if inAmountInLamports else int(inAmount * 10**FROM.decimals)
        paramsQuote = {
            "inputMint":           str(FROM.token_mint),
            "outputMint":          str(TO.token_mint),
            "amount":              finalInAmount,
            "swapMode":            swapParams.swapMode,
            "slippageBps":         swapParams.slippageBps,
            "onlyDirectRoutes":    "true" if swapParams.onlyDirectRoutes    else "false", # For some weird reason Python's `bool` can't be parsed here correctly, looks like JupAg parses it as str
            "asLegacyTransaction": "true" if swapParams.asLegacyTransaction else "false", # For some weird reason Python's `bool` can't be parsed here correctly, looks like JupAg parses it as str
        }
        coinQuote = requests.get(url="https://quote-api.jup.ag/v6/quote", params=paramsQuote).json()

        if "error" in coinQuote:
            if coinQuote["error"] in ["Could not find any route", "The route plan does not consume all the amount, please lower your amount"]:
                logging.warning(f"Swap {str(tokenFrom)} to {str(tokenTo)}: NO ROUTES; bailing...")
                return None
            else:
                logging.warning(f"Swap {str(tokenFrom)} to {str(tokenTo)}: UNKNOWN ERROR; bailing...")
                logging.warning(coinQuote["error"])
                return None

        outLamports = int(coinQuote["outAmount"])
        outAmount   = outLamports / 10**(TO.decimals)

        # Check desired amount, should be at least that
        if desiredOutAmount:
            finalDesiredOutAmountLamports = int(desiredOutAmount) if outAmountInLamports else int(desiredOutAmount * 10**TO.decimals)
            finalDesiredOutAmount         = int(desiredOutAmount) if outAmountInLamports else int(desiredOutAmount * 10**TO.decimals)
            if outLamports < finalDesiredOutAmountLamports:
                logging.warning(f"{tokenTo} desiredOutAmount: {round(finalDesiredOutAmount, 4):.4f} and outAmount: {round(outAmount, 4):.4f}! bailing...")
                return None

        logging.debug(f"Selling {inAmount} of {tokenFrom} for {outAmount} of {tokenTo}...")
        return coinQuote

    # ========================================
    # 
    @staticmethod
    def GetSwapTxBase64(walletAddress: SapysolPubkey,
                        coinQuote:     dict,
                        swapParams:    SapysolJupagParams = SapysolJupagParams()
                       ) -> str:
        paramsSwap = {
            "quoteResponse":             coinQuote,
            "userPublicKey":             str(walletAddress),
            "wrapAndUnwrapSol":          swapParams.wrapAndUnwrapSol,
            "autoMultiplier":            swapParams.quoteAutoMultiplier,     # will 2x of the auto fees
            "dynamicComputeUnitLimit":   swapParams.dynamicComputeUnitLimit, # allow dynamic compute limit instead of max 1,400,000
            "prioritizationFeeLamports": swapParams.quotePrioFeeLamports     # or custom lamports: 1000
        }
        tx = requests.post(url = "https://quote-api.jup.ag/v6/swap", json=paramsSwap).json()
        if not "swapTransaction" in tx:
            logging.warning(f"tx: {tx}")
            logging.warning(f"No swapTransaction in tx! bailing...")
            return None
        return tx["swapTransaction"]

# =============================================================================
# 



#!/usr/bin/python
# =============================================================================
#
#  ######     ###    ########  ##    ##  ######   #######  ##       
# ##    ##   ## ##   ##     ##  ##  ##  ##    ## ##     ## ##       
# ##        ##   ##  ##     ##   ####   ##       ##     ## ##       
#  ######  ##     ## ########     ##     ######  ##     ## ##       
#       ## ######### ##           ##          ## ##     ## ##       
# ##    ## ##     ## ##           ##    ##    ## ##     ## ##       
#  ######  ##     ## ##           ##     ######   #######  ########
#
# =============================================================================
#
# SuperArmor's Python Solana library.
# (c) SuperArmor
#
# module: main
#
# =============================================================================
# 
from sapysol.helpers import EnsurePathExists,       \
                            SetupLogging,           \
                            NestedAttributeExists,  \
                            GetModulePath,          \
                            ListToChunks,           \
                            SapysolPubkey,          \
                            MakePubkey,             \
                            SapysolKeypair,         \
                            MakeKeypair,            \
                            GetFilesFromPath,       \
                            GetKeypairsFromPath,    \
                            GetPubkeysFromKeypairs, \
                            FetchAccount,           \
                            FetchAccounts,          \
                            DivmodJsBignumber

from sapysol.ix import AtaInstruction,              \
                       GetAta,                      \
                       GetOrCreateAtaIx,            \
                       GetTransferTokenIxInternal,  \
                       GetTransferTokenIx,          \
                       WrapSolInstructions,         \
                       UnwrapSolInstruction,        \
                       ComputeBudgetIx,             \
                       ComputePriceIx 

from sapysol.token import SapysolToken

from sapysol.tokenMetadataMetaplex import *
from sapysol.tokenMetadata2022     import *

from sapysol.tx import SapysolTxParams,     \
                       SapysolTxStatus,     \
                       SapysolTxImportMode, \
                       SapysolTx,           \
                       SendAndWaitBatchTx

from sapysol.wallet import SapysolWalletReadonly, \
                           SapysolWallet

from sapysol.jupag import SapysolJupagParams,SapysolJupag

# =============================================================================
# 



#!/usr/bin/python
# =============================================================================
#
#  ######     ###    ########  ##    ##  ######   #######  ##       
# ##    ##   ## ##   ##     ##  ##  ##  ##    ## ##     ## ##       
# ##        ##   ##  ##     ##   ####   ##       ##     ## ##       
#  ######  ##     ## ########     ##     ######  ##     ## ##       
#       ## ######### ##           ##          ## ##     ## ##       
# ##    ## ##     ## ##           ##    ##    ## ##     ## ##       
#  ######  ##     ## ##           ##     ######   #######  ########
#
# =============================================================================
#
# SuperArmor's Python Solana library.
# (c) SuperArmor
#
# module: token
#
# =============================================================================
# 
from   solders.account        import Account
from   solana.rpc.api         import Client, Pubkey, Keypair
from   spl.token.client       import Token
from   spl.token.constants    import ASSOCIATED_TOKEN_PROGRAM_ID, TOKEN_PROGRAM_ID
from   solana.rpc.types       import TxOpts
from   solana.transaction     import Transaction, Signature, Instruction
from   sapysol.helpers        import LAMPORTS_PER_SOL, MakePubkey, MakeKeypair
from   typing                 import List, Any, TypedDict, Union, Optional
from   spl.token.instructions import get_associated_token_address
from   spl.token.core         import AccountInfo, MintInfo, _TokenCore
from  .ix                     import *
from  .tx                     import *
from  .helpers                import MakePubkey, MakeKeypair, NestedAttributeExists, ListToChunks
from  .token_cache            import TokenCacheEntry, TokenCache

import solders.system_program as sp

from solders.rpc.errors import  InvalidParamsMessage
from solders.rpc.responses import RpcKeyedAccountJsonParsed
from solders.account import AccountJSON
from spl.token._layouts import ACCOUNT_LAYOUT, MINT_LAYOUT, MULTISIG_LAYOUT  # type: ignore
import spl.token.instructions as spl_token
import os
import logging


# =============================================================================
# 
class SapysolToken:
    # ========================================
    #
    def __init__(self, connection: Client, tokenMint: SapysolPubkey):
        self.CONNECTION:    Client          = connection
        self.TOKEN_MINT:    Pubkey          = MakePubkey(tokenMint)
        self.TOKEN_INFO:    TokenCacheEntry = TokenCache.GetToken(connection=connection, tokenMint=tokenMint)
        self.TOKEN:         Token           = Token(conn=connection, pubkey=self.TOKEN_MINT, program_id=self.TOKEN_INFO.program_id, payer=None)

    # ========================================
    #
    def AccountExists(self, accountAddress: SapysolPubkey) -> bool:
        pubkey: Pubkey  = MakePubkey(accountAddress)
        result: Account = self.CONNECTION.get_account_info(pubkey=pubkey).value
        return result is not None

    # ========================================
    #
    def GetAccountBalanceLamports(self, accountAddress: SapysolPubkey):
        pubkey: Pubkey = MakePubkey(accountAddress)
        result = self.TOKEN.get_balance(pubkey=pubkey)
        if isinstance(result, InvalidParamsMessage):
            return 0
        return int(result.value.amount)

    # ========================================
    #
    def GetAccountBalance(self, accountAddress: SapysolPubkey) -> float:
        balance: int = self.GetAccountBalanceLamports(accountAddress)
        return balance / 10**self.TOKEN_INFO.decimals

    # ========================================
    #
    def GetWalletAta(self, walletAddress: SapysolPubkey) -> Pubkey:
        pubkey: Pubkey = MakePubkey(walletAddress)
        return get_associated_token_address(owner=pubkey, mint=self.TOKEN_MINT)
    
    # ========================================
    #
    def CreateWalletAtaBatch(self, walletAddresses: List[SapysolPubkey], payer: Union[str, Keypair]) -> List[Pubkey]:
        # Empty transaction is 168 bytes, each ATA creation is 138,
        # limit size is 1232 bytes, that leaves us with up to 7 ATA
        # creations per transaction.
        ixNeeded:      List[Instruction] = []
        resultPubkeys: List[Pubkey]      = []
        for walletAddress in walletAddresses:
            ataIx: AtaInstruction = GetOrCreateAtaIx(connection=self.CONNECTION, tokenMint=self.TOKEN_MINT, owner=MakePubkey(walletAddress))
            resultPubkeys.append(ataIx.pubkey)
            if ataIx.ix:
                ixNeeded.append(ataIx.ix)

        # TODO: change to dynamic size:
        # TODO: if one wallet is sending tokens to multiple wallets, each instruction takes less space

        chunks = ListToChunks(baseList=ixNeeded, chunkSize=7)
        txArray: List[SapysolTx] = []
        for chunk in chunks:
            tx: SapysolTx = SapysolTx(connection=self.CONNECTION, payer=payer)
            tx.FromInstructionsLegacy(instructions=chunk)
            tx.Sign()
        SendAndWaitBatchTx(txArray=txArray)
        return resultPubkeys
    
    # ========================================
    #
    def CreateWalletAta(self, 
                        walletAddress: SapysolPubkey, 
                        payer:         SapysolKeypair) -> Pubkey:
        ataIx: AtaInstruction = GetOrCreateAtaIx(connection=self.CONNECTION, tokenMint=self.TOKEN_MINT, owner=MakePubkey(walletAddress))
        if not ataIx.ix:
            return ataIx.pubkey
        tx: SapysolTx = SapysolTx(connection=self.CONNECTION, payer=payer)
        tx.FromInstructionsLegacy(instructions=ataIx.ix)
        tx.Sign().SendAndWait()
        return ataIx.pubkey

    # ========================================
    #
    def GetWalletAccountAddresses(self, walletAddress: SapysolPubkey) -> List[Pubkey]:
        owner: Pubkey = MakePubkey(walletAddress)
        accounts = self.TOKEN.get_accounts_by_owner_json_parsed(owner=owner)
        account: RpcKeyedAccountJsonParsed
        return [account.pubkey for account in accounts.value]

    # ========================================
    #
    def GetWalletBalanceLamports(self, walletAddress: SapysolPubkey) -> int:
        accountAddress: Pubkey = self.GetWalletAta(walletAddress)
        return self.GetAccountBalanceLamports(accountAddress=accountAddress)

    # ========================================
    #
    def GetWalletBalance(self, walletAddress: SapysolPubkey) -> float:
        accountAddress: Pubkey = self.GetWalletAta(walletAddress)
        return self.GetAccountBalance(accountAddress=accountAddress)

    # ========================================
    #
    def UpdateAuthority(self, newAuthority: Pubkey, payer: SapysolKeypair):
        pass # TODO

    def Mint(self):
        pass # TODO

    def Burn(self):
        pass # TODO

    def Freeze(self):
        pass # TODO

    def Thaw(self):
        pass # TODO

    # ========================================
    #
    def TransferBatch(self):
        pass # TODO

    def Transfer(self):
        pass # TODO

# =============================================================================
# 



#!/usr/bin/python
# =============================================================================
#
#  ######     ###    ########  ##    ##  ######   #######  ##       
# ##    ##   ## ##   ##     ##  ##  ##  ##    ## ##     ## ##       
# ##        ##   ##  ##     ##   ####   ##       ##     ## ##       
#  ######  ##     ## ########     ##     ######  ##     ## ##       
#       ## ######### ##           ##          ## ##     ## ##       
# ##    ## ##     ## ##           ##    ##    ## ##     ## ##       
#  ######  ##     ## ##           ##     ######   #######  ########
#
# =============================================================================
#
# SuperArmor's Python Solana library.
# (c) SuperArmor
#
# module: cache
#
# =============================================================================
# 
from   solders.account        import Account, AccountJSON
from   solders.rpc.errors     import InvalidParamsMessage
from   solders.rpc.responses  import RpcKeyedAccountJsonParsed
from   spl.token.client       import Token
from   spl.token.core         import AccountInfo, MintInfo, _TokenCore
from   solana.rpc.api         import Client, Pubkey, Keypair
from   solana.rpc.types       import TxOpts
from   solana.transaction     import Transaction, Signature, Instruction
from   typing                 import List, Any, TypedDict, Union, Optional
from  .tx                     import *
from  .helpers                import *
from  .ix                     import *
import os
import logging

# =============================================================================
# 
SAPYSOL_TOKEN_VERSION: int = 1

# =============================================================================
# 
class TokenCacheEntry(NamedTuple):
    SAPYSOL_TOKEN_VERSION:  int    #
    token_mint:             Pubkey #
    mint_authority:         Pubkey #
    supply:                 int    #
    decimals:               int    #
    is_initialized:         bool   #
    freeze_authority:       Pubkey #
    program_id:             Pubkey #

class TokenCache:
    # ========================================
    #
    @staticmethod
    def __TokenCachePath() -> str:
        path = os.path.join(os.getenv("HOME"), ".sapysol", "tokens")
        EnsurePathExists(path)
        return path

    @staticmethod
    def __TokenFilename(tokenMint: SapysolPubkey) -> str:
        return os.path.join(TokenCache.__TokenCachePath(), f"{MakePubkey(tokenMint)}.json")

    # ========================================
    #
    @staticmethod
    def __LoadFromFile(tokenMint: SapysolPubkey) -> TokenCacheEntry:
        try:
            tokenCachePath: str = TokenCache.__TokenCachePath()
            tokenInfoFile:  str = TokenCache.__TokenFilename(tokenMint=tokenMint)
            logging.debug(f"Loading token info from file: {tokenInfoFile}")
            if not os.path.isfile(tokenInfoFile):
                return None
            with open(tokenInfoFile) as f:
                tokenInfoJson = json.load(f)
                if "SAPYSOL_TOKEN_VERSION" not in tokenInfoJson:
                    return None
                if tokenInfoJson["SAPYSOL_TOKEN_VERSION"] < SAPYSOL_TOKEN_VERSION:
                    return None

                tokenEntry = TokenCacheEntry(SAPYSOL_TOKEN_VERSION =            tokenInfoJson["SAPYSOL_TOKEN_VERSION"],
                                             token_mint            = MakePubkey(tokenInfoJson["token_mint"]),
                                             mint_authority        = MakePubkey(tokenInfoJson["mint_authority"]),
                                             supply                =            tokenInfoJson["supply"],
                                             decimals              =            tokenInfoJson["decimals"],
                                             is_initialized        =            tokenInfoJson["is_initialized"],
                                             freeze_authority      = MakePubkey(tokenInfoJson["freeze_authority"]),
                                             program_id            = MakePubkey(tokenInfoJson["program_id"]))
                return tokenEntry
        except:
            return None

    # ========================================
    #
    @staticmethod
    def __LoadFromBlockchain(connection: Client, tokenMint: SapysolPubkey) -> TokenCacheEntry:
        tokenCachePath: str = TokenCache.__TokenCachePath()
        tokenInfoFile:  str = TokenCache.__TokenFilename(tokenMint=tokenMint)

        logging.debug(f"Loading token info from Solana Node for token: {str(tokenMint)}")
        accountInfo: Account  = connection.get_account_info(pubkey=MakePubkey(tokenMint)).value
        mintInfo:    MintInfo = Token(conn       = connection, 
                                      pubkey     = MakePubkey(tokenMint),
                                      program_id = accountInfo.owner, 
                                      payer      = None).get_mint_info()
        
        with open(tokenInfoFile, "w") as f:
            tokenMint:   str = str(tokenMint)                  if tokenMint                 else None
            mintAuthStr: str = str(mintInfo.mint_authority)    if mintInfo.mint_authority   else None
            freezeStr:   str = str(mintInfo.freeze_authority)  if mintInfo.freeze_authority else None
            ownerStr:    str = str(accountInfo.owner)          if accountInfo.owner         else None
            json.dump({
                "SAPYSOL_TOKEN_VERSION": SAPYSOL_TOKEN_VERSION,   #
                "token_mint":            str(tokenMint),          #
                "mint_authority":        mintAuthStr,             #
                "supply":                mintInfo.supply,         #
                "decimals":              mintInfo.decimals,       #
                "is_initialized":        mintInfo.is_initialized, #
                "freeze_authority":      freezeStr,               #
                "program_id":            ownerStr,                #
            }, f)
        return TokenCacheEntry(SAPYSOL_TOKEN_VERSION = SAPYSOL_TOKEN_VERSION,
                               token_mint            = tokenMint,
                               mint_authority        = mintInfo.mint_authority,
                               supply                = mintInfo.supply,
                               decimals              = mintInfo.decimals,
                               is_initialized        = mintInfo.is_initialized,
                               freeze_authority      = mintInfo.freeze_authority,
                               program_id            = accountInfo.owner)

    # ========================================
    #
    @staticmethod
    def UpdateTokenCache(connection: Client, tokenMint: SapysolPubkey) -> TokenCacheEntry:
        return TokenCache.__LoadFromBlockchain(connection=connection, tokenMint=tokenMint)

    @staticmethod
    def GetToken(connection: Client, tokenMint: SapysolPubkey) -> TokenCacheEntry:
        tokenInfo = TokenCache.__LoadFromFile(tokenMint=tokenMint)
        return tokenInfo if tokenInfo else TokenCache.__LoadFromBlockchain(connection=connection, tokenMint=tokenMint)

# =============================================================================
# 



#!/usr/bin/python
# =============================================================================
#
#  ######     ###    ########  ##    ##  ######   #######  ##       
# ##    ##   ## ##   ##     ##  ##  ##  ##    ## ##     ## ##       
# ##        ##   ##  ##     ##   ####   ##       ##     ## ##       
#  ######  ##     ## ########     ##     ######  ##     ## ##       
#       ## ######### ##           ##          ## ##     ## ##       
# ##    ## ##     ## ##           ##    ##    ## ##     ## ##       
#  ######  ##     ## ##           ##     ######   #######  ########
#
# =============================================================================
#
# SuperArmor's Python Solana library.
# (c) SuperArmor
#
# module: wallet
#
# =============================================================================
# 
from   solana.rpc.api         import Client, Pubkey, Keypair, Commitment
from   solana.rpc.types       import TxOpts
from   solders.system_program import TransferParams, transfer
from   typing                 import List, Any, TypedDict, Union, Optional
from   dataclasses            import dataclass, field
from   datetime               import datetime
from   enum                   import Enum
from  .helpers                import MakePubkey, MakeKeypair, SapysolKeypair, LAMPORTS_PER_SOL, ListToChunks, SapysolPubkey
from  .tx                     import SapysolTxParams, SapysolTxStatus, SapysolTx, SendAndWaitBatchTx
import base64
import logging
import time

# =============================================================================
#
class SapysolWalletReadonly:
    # ========================================
    #
    def __init__(self, connection: Client, pubkey: SapysolPubkey):
        self.CONNECTION: Client  = connection
        self.PUBKEY:     Pubkey  = MakePubkey(pubkey)

    # ========================================
    #
    def GetBalanceLamports(self) -> int:
        balance = 0
        try: 
            balance = self.CONNECTION.get_balance(self.PUBKEY).value
        except KeyboardInterrupt as e:
            raise
        except:
            pass
        return balance

    # ========================================
    #
    def GetBalanceSol(self) -> float:
        return self.GetBalanceLamports() / LAMPORTS_PER_SOL

    # ========================================
    #


# =============================================================================
#
class SapysolWallet(SapysolWalletReadonly):
    def __init__(self, connection: Client, keypair: SapysolKeypair):
        self.CONNECTION: Client  = connection
        self.KEYPAIR:    Keypair = MakeKeypair(keypair)
        self.PUBKEY:     Pubkey  = self.KEYPAIR.pubkey()

    # ========================================
    #
    def SendLamportsBatch(self, destinationAddresses: List[SapysolPubkey], lamports: int) -> List[SapysolTxStatus]:
        instructions = []
        txArray      = []
        for address in destinationAddresses:
            transferInstruction = transfer(params=TransferParams(from_pubkey=self.KEYPAIR.pubkey(), to_pubkey=MakePubkey(address), lamports=lamports))
            instructions.append(transferInstruction)

        # Empty transaction is 168 bytes, each SOL transfer is 49,
        # limit size is 1232 bytes, that leaves us with up to 21 SOL 
        # transfers per transaction.
        instructionsChunked = ListToChunks(baseList=instructions, chunkSize=20)
        for chunk in instructionsChunked:
            tx = SapysolTx(connection=self.CONNECTION, payer=self.KEYPAIR)
            tx.FromInstructionsLegacy(chunk)
            tx.Sign()
            txArray.append(tx)
        return SendAndWaitBatchTx(txArray=txArray)

    # ========================================
    #
    def SendLamports(self, destinationAddress: SapysolPubkey, lamports: int) -> SapysolTxStatus:
        return self.SendLamportsBatch(destinationAddresses=[destinationAddress], lamports=lamports)[0]

    # ========================================
    #
    def SendLamportsAll(self, destinationAddress: SapysolPubkey) -> SapysolTxStatus:
        feeLamports:     int = 5000
        balanceLamports: int = self.GetBalanceLamports()
        if balanceLamports <= feeLamports:
            return SapysolTxStatus.SUCCESS

        return self.SendLamports(destinationAddress=destinationAddress, lamports=balanceLamports-feeLamports)

    # ========================================
    #
    def SendSolBatch(self, destinationAddresses: List[SapysolPubkey], amountSol: float) -> List[SapysolTxStatus]:
        return self.SendLamportsBatch(destinationAddresses=destinationAddresses, lamports=int(amountSol*LAMPORTS_PER_SOL))

    def SendSol(self, destinationAddress: SapysolPubkey, amountSol: float) -> SapysolTxStatus:
        return self.SendLamportsBatch(destinationAddresses=[destinationAddress], lamports=int(amountSol*LAMPORTS_PER_SOL))[0]

# =============================================================================
#



#!/usr/bin/python
# =============================================================================
#
#  ######     ###    ########  ##    ##  ######   #######  ##       
# ##    ##   ## ##   ##     ##  ##  ##  ##    ## ##     ## ##       
# ##        ##   ##  ##     ##   ####   ##       ##     ## ##       
#  ######  ##     ## ########     ##     ######  ##     ## ##       
#       ## ######### ##           ##          ## ##     ## ##       
# ##    ## ##     ## ##           ##    ##    ## ##     ## ##       
#  ######  ##     ## ##           ##     ######   #######  ########
#
# =============================================================================
#
# SuperArmor's Python Solana library.
# (c) SuperArmor
#
# module: tx
#
# =============================================================================
# 
from   solana.rpc.api                       import Client, Pubkey, Keypair, Commitment
from   solana.rpc.types                     import TxOpts
from   solana.transaction                   import Transaction, Signature, Instruction
from   solders.address_lookup_table_account import AddressLookupTableAccount
from   solders.message                      import to_bytes_versioned, MessageV0
from   solders.transaction                  import VersionedTransaction, Signer
from   solders.transaction_status           import EncodedTransactionWithStatusMeta
from   typing                               import List, Any, TypedDict, Union, Optional, Literal
from   dataclasses                          import dataclass, field
from   datetime                             import datetime
from   enum                                 import Enum
from  .helpers                              import MakeKeypair, SapysolKeypair, NestedAttributeExists
import base64
import logging
import time

# ================================================================================
#
@dataclass
class SapysolTxParams:
    maxSecondsPerTx:       int        = 30          # 
    sleepBetweenRetry:     float      = 0.3         # 
    skipConfirmation:      bool       = True        # 
    skipPreFlight:         bool       = True        # 
    maxRetries:            int        = 0           # 
    blockhashCommitment:   Commitment = "finalized" # 
    transactionCommitment: Commitment = "confirmed" # 

# ================================================================================
#
class SapysolTxStatus(Enum):
    PENDING = 1 # Pending transaction is not (yet) processed
    TIMEOUT = 2 # We couldn't wait till transaction gets processed, it MAY be processed
    FAIL    = 3 # 100% Fail
    SUCCESS = 4 # 100% Success

SapysolTxImportMode = Literal["auto", "legacy", "versioned"]

# ================================================================================
#
class SapysolTx:
    def __init__(self, connection: Client, payer: SapysolKeypair, txParams: SapysolTxParams = SapysolTxParams()):

        self.CONNECTION:       Client                                   = connection
        self.PAYER:            Keypair                                  =  MakeKeypair(payer)
        self.SIGNERS:          List[Signer]                             = [MakeKeypair(payer)]
        self.TX_PARAMS:        SapysolTxParams                          = txParams
        self.CONFIRMED_TX:     EncodedTransactionWithStatusMeta         = None
        self.CONFIRMED_RESULT: SapysolTxStatus                          = SapysolTxStatus.PENDING
        self.RAW_TX:           Union[VersionedTransaction, Transaction] = None
        self.SENT_DT:          datetime                                 = None
        self.TXID:             Signature                                = None
        self.LAST_VALID_BLOCKHEIGHT: int                                = None

    # ========================================
    #
    def FromInstructionsLegacy(self, 
                               instructions: List[Instruction],
                               signers:      List[Signer] = None) -> "SapysolTx":
        if signers:
            self.SIGNERS = signers
        latestBlockHash = self.CONNECTION.get_latest_blockhash(commitment=self.TX_PARAMS.blockhashCommitment).value
        self.RAW_TX     = Transaction(recent_blockhash = latestBlockHash.blockhash,
                                      instructions     = instructions)

        return self

    # ========================================
    #
    def FromInstructionsVersioned(self, 
                                  instructions:        List[Instruction],
                                  signers:             List[Signer] = None,
                                  lookupTableAccounts: List[AddressLookupTableAccount] = []) -> "SapysolTx":
        if signers:
            self.SIGNERS = signers
        latestBlockHash = self.CONNECTION.get_latest_blockhash(commitment=self.TX_PARAMS.blockhashCommitment).value
        msg = MessageV0.try_compile(
            payer            = self.PAYER.pubkey(),
            instructions     = instructions,
            recent_blockhash = latestBlockHash.blockhash,
            address_lookup_table_accounts = lookupTableAccounts,
        )
        self.RAW_TX = VersionedTransaction(msg, signers if signers else [self.PAYER])
        return self

    # ========================================
    #
    def FromBytes(self, b: bytes, importMode: SapysolTxImportMode = "auto") -> "SapysolTx":
        match importMode:
            case "auto":
                try:
                    return self.FromBytes(b=b, importMode="versioned")
                except:
                    return self.FromBytes(b=b, importMode="legacy")
            case "legacy":
                self.RAW_TX = Transaction.deserialize(b)
            case "versioned":
                self.RAW_TX = VersionedTransaction.from_bytes(b)
        return self

    # ========================================
    #
    def FromBase64(self, b64: str, importMode: SapysolTxImportMode = "auto") -> "SapysolTx":
        b = base64.b64decode(b64)
        return self.FromBytes(b=b, importMode=importMode)

    # ========================================
    #
    def Encode(self) -> str:
        encodedTx: str = base64.b64encode(bytes(self.RAW_TX)).decode("utf-8")
        return encodedTx

    # ========================================
    #
    def Decode(self) -> bytes:
        if isinstance(self.RAW_TX, VersionedTransaction):
            return bytes(self.RAW_TX)
        elif isinstance(self.RAW_TX, Transaction):
            return self.RAW_TX.serialize()

    # ========================================
    #
    def SetTxParams(self, txParams: SapysolTxParams) -> bytes:
        self.TX_PARAMS: SapysolTxParams = txParams

    # ========================================
    # 
    def Sign(self, signersOverride: List[SapysolKeypair] = None) -> "SapysolTx":
        signers = self.SIGNERS if signersOverride is None else [MakeKeypair(signer) for signer in signersOverride]
        if signers is None:
            signers = [self.PAYER]
        if signers is None:
            raise(Exception(f"SapysolTx::Sign(): no signers specified!"))

        if isinstance(self.RAW_TX, VersionedTransaction):
            self.RAW_TX = VersionedTransaction(message  = self.RAW_TX.message,
                                               keypairs = signers)
        elif isinstance(self.RAW_TX, Transaction):
            self.RAW_TX.fee_payer = signers[0].pubkey()
            self.RAW_TX.sign(*signers)

        return self

    # ========================================
    #
    def __SendInternal(self, connection: Client, txParams: SapysolTxParams) -> "SapysolTx":
        if self.CONFIRMED_RESULT != SapysolTxStatus.PENDING:
            return self

        if self.SENT_DT is None:
            self.SENT_DT = datetime.now()

        if self.LAST_VALID_BLOCKHEIGHT is None:
            latestBlockHash = (connection.get_latest_blockhash(commitment=txParams.blockhashCommitment)).value
            self.LAST_VALID_BLOCKHEIGHT: int = latestBlockHash.last_valid_block_height

        if txParams.maxSecondsPerTx is not None and (datetime.now() - self.SENT_DT).seconds >= txParams.maxSecondsPerTx:
            self.CONFIRMED_RESULT = SapysolTxStatus.TIMEOUT
            logging.info(f"{self.CONFIRMED_RESULT.name}: https://solscan.io/tx/{self.TXID}")
            return self

        blockheight = (connection.get_block_height()).value
        if blockheight < self.LAST_VALID_BLOCKHEIGHT:
            logging.debug(f"SapysolTx::Send() blockheight={blockheight}; lastValidBlockHeight={self.LAST_VALID_BLOCKHEIGHT}; {self.LAST_VALID_BLOCKHEIGHT-blockheight}")

            txOpts: TxOpts = TxOpts(skip_confirmation = txParams.skipConfirmation, 
                                    skip_preflight    = txParams.skipPreFlight, 
                                    max_retries       = txParams.maxRetries)
            self.TXID: Signature = (connection.send_raw_transaction(txn=self.Decode(), opts=txOpts)).value

        return self

    # ========================================
    # Can send to one or many endpoints at the same time.
    # TODO: add description
    #
    def Send(self,
             connectionOverride: Union[str, Client, List[Client]] = None,
             txParamsOverride:   SapysolTxParams = None) -> "SapysolTx":

        useDefaultConnection:   bool = connectionOverride is None
        useSingleOverride:      bool = isinstance(connectionOverride, Client)
        useSingleOverrideStr:   bool = isinstance(connectionOverride, str)
        useMultipleOverride:    bool = isinstance(connectionOverride, list) and all(isinstance(n, Client) for n in connectionOverride)
        useMultipleOverrideStr: bool = isinstance(connectionOverride, list) and all(isinstance(n, str)    for n in connectionOverride)
        txParams: SapysolTxParams    = txParamsOverride if txParamsOverride else self.TX_PARAMS

        # None is given
        if useDefaultConnection:
            self.__SendInternal(connection=self.CONNECTION, txParams=txParams)

        # `Client` is given
        elif useSingleOverride:
            self.__SendInternal(connection=connectionOverride, txParams=txParams)

        # connection string is given
        elif useSingleOverrideStr:
            self.__SendInternal(connection=Client(connectionOverride), txParams=txParams)

        # multiple `Client`s are given
        elif useMultipleOverride:
            for connection in connectionOverride:
                self.__SendInternal(connection=connection, txParams=txParams)

        # multiple connection strings are given
        elif useMultipleOverrideStr:
            for connection in connectionOverride:
                self.__SendInternal(Client(connection), txParams=txParams)

        # ?????
        else:
            raise Exception("SapysolTx::Send() Error! `connectionOverride` has invalid type!")

        return self

    # ========================================
    #
    def __ConfirmInternal(self, connection: Client) -> SapysolTxStatus:
        if self.CONFIRMED_RESULT != SapysolTxStatus.PENDING:
            return self.CONFIRMED_RESULT

        if self.TXID is None:
            return SapysolTxStatus.PENDING

        if self.CONFIRMED_TX is not None:
            self.CONFIRMED_RESULT = SapysolTxStatus.SUCCESS if self.CONFIRMED_TX.meta.err is None \
                               else SapysolTxStatus.FAIL
            logging.info(f"{self.CONFIRMED_RESULT.name}: https://solscan.io/tx/{self.TXID}")
            return self.CONFIRMED_RESULT

        try:
            trans = connection.get_transaction(tx_sig     = self.TXID,
                                               commitment = self.TX_PARAMS.transactionCommitment,
                                               max_supported_transaction_version=0)

            if NestedAttributeExists(target=trans, attributePath="value.transaction"):
                self.CONFIRMED_TX = trans.value.transaction
                return self.Confirm() # one more pass because we need to hit `if self.CONFIRMED_TX is not None` case

        except KeyboardInterrupt as e:
            raise
        except Exception as e:
            logging.error(e, exc_info=(type(e), e, e.__traceback__))

        return self.CONFIRMED_RESULT

    # ========================================
    #
    def Confirm(self) -> SapysolTxStatus:
        return self.__ConfirmInternal(connection=self.CONNECTION)

    # ========================================
    # TODO: for connection list send all transactions threaded
    #
    def SendAndWait(self, connectionOverride: Union[Client, List[Client]] = None) -> SapysolTxStatus:
        while True:
            r: SapysolTxStatus = (self.Send(connectionOverride=connectionOverride)).Confirm()
            if r != SapysolTxStatus.PENDING:
                return r

            if self.TX_PARAMS.sleepBetweenRetry and self.TX_PARAMS.sleepBetweenRetry > 0:
                time.sleep(self.TX_PARAMS.sleepBetweenRetry)

# ================================================================================
# TODO: return not only result but full tx info
def SendAndWaitBatchTx(txArray:  List[SapysolTx],
                       txParams: SapysolTxParams = SapysolTxParams()) -> List[SapysolTxStatus]:
    while True:
        txNum = 0
        results: List[SapysolTxStatus] = []

        tx: SapysolTx
        for tx in txArray:
            r: SapysolTxStatus = tx.Send().Confirm()
            results.append(r)
            if r != SapysolTxStatus.PENDING:
                txNum += 1

        if txNum >= len(txArray):
            return results

        if txParams.sleepBetweenRetry and txParams.sleepBetweenRetry > 0:
            time.sleep(txParams.sleepBetweenRetry)

# ================================================================================
#



#!/usr/bin/python
# =============================================================================
#
#  ######     ###    ########  ##    ##  ######   #######  ##       
# ##    ##   ## ##   ##     ##  ##  ##  ##    ## ##     ## ##       
# ##        ##   ##  ##     ##   ####   ##       ##     ## ##       
#  ######  ##     ## ########     ##     ######  ##     ## ##       
#       ## ######### ##           ##          ## ##     ## ##       
# ##    ## ##     ## ##           ##    ##    ## ##     ## ##       
#  ######  ##     ## ##           ##     ######   #######  ########
#
# =============================================================================
#
# SuperArmor's Python Solana library.
# (c) SuperArmor
#
# module: mass token sell (dump) via Jupiter
#
# =============================================================================
# 
from   typing            import List, Union, Any, Literal
from   queue             import Queue, Empty
from   threading         import Thread
from   solana.exceptions import SolanaRpcException
import copy
import time
import logging

# =============================================================================
# 
SAPYSOL_ERROR_ACTION = Literal["ignore", "print", "raise"]

# =============================================================================
# 
class SapysolBatcher:
    def __init__(self, 
                 callback:    Any,
                 entityList:  List[Any],
                 entityKwarg: str = None,
                 numThreads:  int = 10,
                 args             = (),
                 kwargs           = {}):

        self.ENTITY_LIST:  List[Any]    = entityList
        self.ENTITY_KWARG: str          = entityKwarg
        self.CALLBACK                   = callback
        self.ARGS                       = copy.deepcopy(args)
        self.KWARGS                     = copy.deepcopy(kwargs)
        self.NUM_THREADS:  int          = numThreads
        self.QUEUE:        Queue        = Queue()
        self.THREADS:      List[Thread] = []
        for entity in entityList:
            self.QUEUE.put(entity)
        self.RPC_ERROR_ACTION: SAPYSOL_ERROR_ACTION = "ignore"
        self.ALL_ERROR_ACTION: SAPYSOL_ERROR_ACTION = "ignore"

    # ========================================
    #
    def Start(self, 
              sleepTime: float = 0.1, 
              rpcErrorAction: SAPYSOL_ERROR_ACTION = "print",
              allErrorAction: SAPYSOL_ERROR_ACTION = "print") -> None:
        self.RPC_ERROR_ACTION: SAPYSOL_ERROR_ACTION = rpcErrorAction if rpcErrorAction else "print"
        self.ALL_ERROR_ACTION: SAPYSOL_ERROR_ACTION = allErrorAction if allErrorAction else "print"

        for _ in range(self.NUM_THREADS):
            thr = Thread(target=self.__ProcessSingle)
            thr.start()
            self.THREADS.append(thr)

        while not self.IsDone():
            time.sleep(sleepTime)

    # ========================================
    #
    def IsDone(self) -> bool:
        thr: Thread
        for thr in self.THREADS:
            if thr.is_alive():
                return False
        return True

    # ========================================
    #
    def __ProcessSingle(self):
        def __SingleCall(entity):
            while True:
                try:
                    # Check if entity has kwarg name
                    if self.ENTITY_KWARG:
                        self.KWARGS[self.ENTITY_KWARG] = entity
                        self.CALLBACK(*self.ARGS, **self.KWARGS)
                    else:
                        self.CALLBACK(entity, *self.ARGS, **self.KWARGS)
                    return
                except SolanaRpcException as e:
                    match self.RPC_ERROR_ACTION:
                        case "ignore":
                            pass
                        case "print":
                            logging.error(f"SapysolBatcher::__ProcessSingle(), RPC error:\n{e}")
                        case "raise":
                            raise
                except Exception as e:
                    match self.ALL_ERROR_ACTION:
                        case "ignore":
                            pass
                        case "print":
                            logging.error(f"SapysolBatcher::__ProcessSingle(), Error:\n{e}")
                        case "raise":
                            raise

        while True:
            try:
                if self.QUEUE.qsize() <= 0:
                    return
                entity = self.QUEUE.get()
                self.QUEUE.task_done()
                __SingleCall(entity=entity)
            except Empty:
                return
            except:
               raise

# =============================================================================
# 


#!/usr/bin/python
# =============================================================================
#
#  ######     ###    ########  ##    ##  ######   #######  ##       
# ##    ##   ## ##   ##     ##  ##  ##  ##    ## ##     ## ##       
# ##        ##   ##  ##     ##   ####   ##       ##     ## ##       
#  ######  ##     ## ########     ##     ######  ##     ## ##       
#       ## ######### ##           ##          ## ##     ## ##       
# ##    ## ##     ## ##           ##    ##    ## ##     ## ##       
#  ######  ##     ## ##           ##     ######   #######  ########
#
# =============================================================================
#
# SuperArmor's Python Solana library.
# (c) SuperArmor
#
# module: instructions
#
# =============================================================================
# 
from   solana.rpc.api         import Client, Pubkey, Keypair
from   solana.transaction     import Instruction
from   solders.system_program import transfer
from   solders.compute_budget import set_compute_unit_limit, set_compute_unit_price
from   spl.token.constants    import WRAPPED_SOL_MINT as NATIVE_MINT, TOKEN_PROGRAM_ID
from   spl.token.instructions import get_associated_token_address, create_associated_token_account, close_account, sync_native, CloseAccountParams, SyncNativeParams
import spl.token.instructions as     splToken
from   typing                 import List, Any, TypedDict, Union, Optional, NamedTuple
from  .helpers                import MakePubkey, MakeKeypair, SapysolPubkey
from  .token_cache            import TokenCacheEntry, TokenCache

# ===============================================================================
# Creating ATA takes 138 bytes per instruction.
# If the same sender address is used, each next transfer takes 74 bytes per instruction.
# But also may vary, because it depends.
#
class AtaInstruction(NamedTuple):
    pubkey: Pubkey
    ix:     Instruction = None

def GetAta(tokenMint: SapysolPubkey, owner: SapysolPubkey) -> Pubkey:
    return get_associated_token_address(owner=MakePubkey(owner), mint=MakePubkey(tokenMint))

def GetOrCreateAtaIx(connection: Client,
                     tokenMint:  SapysolPubkey,
                     owner:      SapysolPubkey,
                     payer:      SapysolPubkey = None,
                     allowOwnerOffCurve: bool = True) -> AtaInstruction:

    _tokenMint = MakePubkey(tokenMint)
    _owner     = MakePubkey(owner)
    _payer     = MakePubkey(payer)
    _ix        = None

    if allowOwnerOffCurve == False and not owner.is_on_curve():
        raise("GetOrCreateATAInstruction(): allowOwnerOffCurve = False, but `owner` is off curve address!")

    if _payer is None:
        _payer = _owner
    ataAddress = GetAta(owner=owner, tokenMint=tokenMint)
    account    = connection.get_account_info(ataAddress)
    if account.value is None:
        _ix = create_associated_token_account(payer=_payer, owner=_owner, mint=_tokenMint)
    return AtaInstruction(pubkey=ataAddress, ix=_ix)

# ===============================================================================
# `transfer` is deprecated, using `transfer_checked`.
# Transfering SPL tokens takes 114 bytes per instruction.
# If the same sender address is used, each next transfer takes 50 bytes per instruction.
# But also may vary, because it depends.
#
def GetTransferTokenIxInternal(tokenProgramID: SapysolPubkey,
                               tokenMint:      SapysolPubkey,
                               decimals:       int,
                               senderWallet:   SapysolPubkey,
                               senderAta:      SapysolPubkey,
                               receiverAta:    SapysolPubkey,
                               amountLamports: int) -> Instruction:
    return splToken.transfer_checked(
        splToken.TransferCheckedParams(
            program_id = MakePubkey(tokenProgramID),
            source     = MakePubkey(senderAta),
            mint       = MakePubkey(tokenMint),
            dest       = MakePubkey(receiverAta),
            owner      = MakePubkey(senderWallet),
            amount     = amountLamports,
            decimals   = decimals,
            signers    = [MakePubkey(senderWallet)],
        )
    )

# ===============================================================================
#
def GetTransferTokenIx(connection:       Client,
                       tokenMint:        SapysolPubkey,
                       senderWallet:     SapysolPubkey,
                       receiverWallet:   SapysolPubkey,
                       amount:           int,
                       amountIsLamports: bool = True,
                       allowCreateAta:   bool = True) -> List[Instruction]:
    result: List[Instruction] = []
    if allowCreateAta:
        ataIx: AtaInstruction = GetOrCreateAtaIx(connection = connection,
                                                tokenMint  = tokenMint,
                                                owner      = receiverWallet,
                                                payer      = senderWallet)
        if ataIx.ix:
            result.append(ataIx.ix)

    tokenInfo:    TokenCacheEntry = TokenCache.GetToken(connection=connection, tokenMint=tokenMint)
    sendLamports: int             = amount if amountIsLamports else (int(amount * 10**tokenInfo.decimals))
    transferIx = GetTransferTokenIxInternal(tokenProgramID = tokenInfo.program_id,
                                            tokenMint      = tokenMint,
                                            decimals       = tokenInfo.decimals,
                                            senderWallet   = senderWallet,
                                            senderAta      = get_associated_token_address(owner=MakePubkey(senderWallet),   mint=MakePubkey(tokenMint)),
                                            receiverAta    = get_associated_token_address(owner=MakePubkey(receiverWallet), mint=MakePubkey(tokenMint)),
                                            amountLamports = sendLamports)
    result.append(transferIx)
    return result

# ===============================================================================
#
def WrapSolInstructions(source:   SapysolPubkey,
                        dest:     SapysolPubkey,
                        lamports: int) -> List[Instruction]:

    _source = MakePubkey(source) # preserve original `source`
    _dest   = MakePubkey(dest)   # preserve original `dest`
    return [
        transfer({
            "from_pubkey": _source,
            "to_pubkey":   _dest,
            "lamports":    lamports
        }),
        sync_native(SyncNativeParams(program_id=TOKEN_PROGRAM_ID, account=_dest))
    ]

# ===============================================================================
# 
def UnwrapSolInstruction(owner:              SapysolPubkey,
                         allowOwnerOffCurve: bool = True) -> Instruction:

    _owner = MakePubkey(owner) # preserve original `owner`
    if allowOwnerOffCurve == False and not _owner.is_on_curve():
        raise("UnwrapSolInstruction(): allowOwnerOffCurve = False, but `owner` is off curve address!")

    WSolAtaAccount = get_associated_token_address(owner=_owner, mint=NATIVE_MINT)
    if WSolAtaAccount:
        return close_account(params=CloseAccountParams(program_id = TOKEN_PROGRAM_ID,
                                                       account    = WSolAtaAccount,
                                                       dest       = _owner,
                                                       owner      = _owner))
    return None

# ===============================================================================
#
def ComputeBudgetIx(units=1_400_000) -> Instruction:
    return set_compute_unit_limit(units=units)

def ComputePriceIx(microLamports=1) -> Instruction:
    return set_compute_unit_price(micro_lamports=microLamports)

# ===============================================================================
#
